#ifndef _yahtzee_H_
#define _yahtzee_H_

#include "config.h"

#define MAX_NUMBER_OF_PLAYERS 6
#define MAX_NAME_LENGTH 8
#define NUM_UPPER 6
#define NUM_LOWER 7
#define NUM_FIELDS (NUM_UPPER + NUM_LOWER)

#define H_3 6
#define H_4 7
#define H_FH 8
#define H_SS 9
#define H_LS 10
#define H_YA 11
#define H_CH 12

typedef struct
{
	char name[MAX_NAME_LENGTH + 1];
	short used[NUM_FIELDS];
	int score[NUM_FIELDS];
	int finished;
	int comp;
} Player;

#endif /* _yahtzee_H_ */
