#ifndef _config_H_
#define _config_H_

#define SCOREDIR "/usr/local/lib"
#define SCOREFNAME "yahtzee.sco"	/* must allow .L extension */

#define NUM_TOP_PLAYERS 10
#define COMPUTER_DELAY 1		/* sec */

#if 0
/*
   your linux either:
   a) has no rename, or
   b) can't rename regular files if newname exists
*/
#define HAS_RENAME			/* comment out if you don't */
#endif

#endif /* _config_H_ */
